syntax = "proto3";
package grpc.gateway.runtime;
option go_package = "internal";

// StreamError is a response type which is returned when
// streaming rpc returns an error.
message StreamError {
	int32 grpc_code = 1;
	int32 http_code = 2;
	string message = 3;
	string http_status = 4;
}
